/*
  $Id: swu-lock.c 37 2007-05-02 14:41:02Z carlo $

  swu-lock: simple Open Lock Daemon client.

  Copyright (c) 2006 Carlo Strozzi

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 dated June, 1991.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Note:

  this program links libraries distributed with the Open Lock Daemon,
  written by Alberto Bertogli (albertogli@telpin.com.ar). According
  to the doc/license file distributed with that program:

    'The act of linking/importing the library into another program
     does NOT in itself make that program considered a derivative
     work of this Original Work.' 

  According to that, it sounds ok if I distribute swu-lock.c under
  the GNU GPL as opposed to the Open Software License that covers
  the Open Lock Daemon code.

*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

#include "compiler.h"
#include "old.h"

static const char *progname = "swu-lock";  /* global pointer to argv[0] */

/* global vars */
static int fd, argcount, verbose, mylocks, child=0;
static char **arglist;
static char salt[32] = "SWU";		   /* default salt value */

static int eusage(void) {
       fprintf(stderr,\
	 "Usage: %s [-v|--verbose] [-t|--timeout t] host resource [resource ...]\n",\
	 progname);
       return 2;
}

static void cleanup() {

     int i, rv=0;
     char obj[1024];

     /* Do not release the locks if child! Some of those locks could
	in the meantime have already been explicitly released by the
	calling program through swu-unlock, and acquired again by
	some other process. If we release them we are at risk of
	removing locks that do no longer belong to us! On the other
	hand, if we do not release them and if for any reason our
	caller died before releasing them explicitly, a new process
	traying to acquire them will simply have to wait for them to
	be timed-out by the daemon, which is a matter of a few seconds
	at most. */

     if (child) return;

     if (verbose)
	fprintf(stderr,"%s: clearing any acquired locks ... \n",progname);

     for (i = 0; i < mylocks; i++) {
         snprintf(obj, 1023, "%s:%s", salt, arglist[i]);

	 if (verbose) fprintf(stderr,\
	    "%s: unlocking resource '%s' ... ",progname,arglist[i]);

         rv = old_unlock(fd, obj);

         if (unlikely(rv < 0)) {
	    if (verbose) fprintf(stderr,"failed!\n");
	    exit(1);
	 }

	 if (verbose) {
	    if (rv) fprintf(stderr,"ok.\n");
	    else fprintf(stderr,"it wasn't locked!\n");
	 }
     }

     return;
}

static void onintr() {
     if (verbose)
	fprintf(stderr,"%s: pid %d interrupted!\n",progname,getpid());
     cleanup();
     exit(1);		/* killed or something */
}

static void onalarm() {
     fprintf(stderr,"%s: time-out alarm received, exiting.\n",progname);
     cleanup();
     exit(1);		/* timed-out */
}

int main(int argc, char **argv) {

    int rv=0, i=0, t=0, min_args=3;
    char obj[1024];
    pid_t pid;
    struct stat statbuf;
    char *semaphore = "";

    /* the lock password is passed through env, not to
       expose it as a command-line argument. */

    if (getenv("SWU_LOCK_SALT") != NULL)
	 strncpy(salt,getenv("SWU_LOCK_SALT"),31);

    while (++i < argc && *argv[i] == '-') {

       min_args++;

       if (!strcmp(argv[i], "-t") ||
	   !strcmp(argv[i], "--timeout")) {

	 if (++i >= argc || *argv[i] == '-') exit(eusage());

	 min_args++;

	 t = atoi(argv[i]);
       }

       else if (!strcmp(argv[i], "-v") ||
	   !strcmp(argv[i], "--verbose")) verbose = 1;

       if (!strcmp(argv[i], "-s") ||
	   !strcmp(argv[i], "--semaphore")) {

	 if (++i >= argc || *argv[i] == '-') exit(eusage());

	 min_args++;

	 semaphore = argv[i];
       }

    }

    if (argc < min_args) exit(eusage());

    argv += (min_args - 2);
    argc -= (min_args - 2);

    /* now argv[0] is the hostname */

    fd = old_connect(argv[0], 2626);

    if (fd < 0) {

       /* If the daemon is supposed to sit on the local host, try and
	  start it (assume it is in PATH), before giving up. */

       if (!strcmp(argv[0], "localhost")) {

	  system("old");

	  sleep(1);		/* allow for daemon initialization */

	  fd = old_connect(argv[0], 2626);	/* try again ... */
       }
    }

    if (fd < 0) {			/* no luck, give up */

       fprintf(stderr,\
	   "%s: could not connect to daemon on '%s'\n",progname,argv[0]);

       return 1;
    }

    argv += 1;
    argc -= 1;

    /* make argv[] and argc globally visible */

    argcount = argc;
    arglist = argv;

    /* catch the usual signals 1 2 3 13 15 */

    if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
			   signal(SIGHUP, onintr);	/* signal  1 */

    if (signal(SIGINT, SIG_IGN) != SIG_IGN)
			   signal(SIGINT, onintr);	/* signal  2 */

    if (signal(SIGQUIT, SIG_IGN) != SIG_IGN)
			   signal(SIGQUIT, onintr);	/* signal  3 */

    if (signal(SIGPIPE, SIG_IGN) != SIG_IGN)
			   signal(SIGPIPE, onintr);	/* signal 13 */

    if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
			   signal(SIGTERM, onintr);	/* signal 15 */

    if (signal(SIGALRM, SIG_IGN) != SIG_IGN)
			    signal(SIGALRM, onalarm);	/* signal 14 */

    if (verbose && t)
	fprintf(stderr,"%s: scheduling a %d-second timeout ... ",progname,t);

    alarm(t);

    if (verbose && t) fprintf(stderr,"scheduled.\n");

    for (mylocks = 0; mylocks < argc; mylocks++) {

	snprintf(obj, 1023, "%s:%s", salt, argv[mylocks]);
		
	/* sleep(4);          testing */

	if (verbose) fprintf(stderr,\
	    "%s: locking resource '%s' ... ",progname,argv[mylocks]);

	rv = old_lock(fd, obj);

	if (unlikely(rv < 1)) {
	   if (verbose) fprintf(stderr,"failed!\n");
	   cleanup();
	   exit(1);
	}

	if (verbose) fprintf(stderr,"ok.\n");
    }

    if ((pid = fork()) < 0) {
       perror("fork");
       exit(1);
    }

    if (pid == 0) {					/* child  */

       child = 1;

       /*
	  The sole purpose of this child is to keep the connection
	  with old(1) open for long enough for all practical purposes.
	  It will be up to the calling program to kill this child earlier
	  if necessary, in which case the acquired locks will be cleared
	  by onintr().
       */

       /* catch the usual signals 1 2 3 13 15 */

       if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
			   signal(SIGHUP, onintr);	/* signal  1 */

       if (signal(SIGINT, SIG_IGN) != SIG_IGN)
			   signal(SIGINT, onintr);	/* signal  2 */

       if (signal(SIGQUIT, SIG_IGN) != SIG_IGN)
			   signal(SIGQUIT, onintr);	/* signal  3 */

       if (signal(SIGPIPE, SIG_IGN) != SIG_IGN)
			   signal(SIGPIPE, onintr);	/* signal 13 */

       if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
			   signal(SIGTERM, onintr);	/* signal 15 */

       if (signal(SIGALRM, SIG_IGN) != SIG_IGN)
			    signal(SIGALRM, onalarm);	/* signal 14 */

       if (verbose)
	  fprintf(stderr,"%s child: pid %d\n",progname,getpid());

       if (*semaphore == '\0') {
	  sleep(60);			/* keep the connection open */
	  cleanup();
	  exit(0);			/* quit child after timeout */
       }

       if (verbose) fprintf(stderr,\
		"%s child: scheduling a 60-second timeout ... ",progname);

       /* quit child either when the semaphore file vanishes or
	  when the alarm is triggered, whichever comes first. */

       alarm(60);

       if (verbose) fprintf(stderr,"scheduled.\n");

       while (*semaphore) {
	  if (stat(semaphore,&statbuf) < 0) {
	     cleanup();
	     exit(0);
	  }
	  usleep(1000);
       }
    }					/* end of child */

    exit(0);				/* quit parent */
}

/* EOF */
